/*
  ikkeiplexing.cpp
	 - Using MsTimer2() for ATMega328P
	 - Using Timer interrupt for AVR64DD28
  by ikkei <ikkei@zeus.eonet.ne.jp> http://blog.goo.ne.jp/jh3kxm
  
  
  Original:
  Charlieplexing.cpp
  Alex Wenger <a.wenger@gmx.de> http://arduinobuch.wordpress.com/
  Matt Mets <mahto@cibomahto.com> http://cibomahto.com/
  
  Timer init code from MsTimer2 - Javier Valencia <javiervalencia80@gmail.com>
  Misc functions from Benjamin Sonnatg <benjamin@sonntag.fr>
  
  History:
    2009-12-30 - V0.0 wrote the first version at 26C3/Berlin
    2010-01-01 - V0.1 adding misc utility functions 
      (Clear, Vertical,  Horizontal) comment are Doxygen complaints now
    2010-05-27 - V0.2 add double-buffer mode
    
    2010-11-24
      Modified for ikkeiplexing shield (direct two 8x8 dot matrix LED shield)
      With brightness equalization
	  by ikkei <ikkei@zeus.eonet.ne.jp> http://blog.goo.ne.jp/jh3kxm
	2012-12-23
	  Modified for Titchaino with switch
	  by ikkei
	2015-03-16
	  Modified for Titchaino_uni with switch
      by ikkei
	2018-07-13
	  Rename ikkeiplexing.cpp from Charlieplexing.cpp
	  by ikkei
	2024-04-30
	  Applied to AVR64DD28.
	  Rename Dominoduino uni from Titchaino_uni
      by ikkei

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <Arduino.h>
#include <avr/interrupt.h>
#include "ikkeiplexing.h"

#ifdef __AVR_ATmega328P__
  #include <MsTimer2.h>
#endif


const byte maptable[][16] PROGMEM = {
	{  1,  4, 15,  7, 16,  8, 11, 12,  6, 10, 17,  2, 14,  3,  9,  5},	// IKKEIPLEXING		for ikkeiplexing shield
	{ 17, 13,  9, 11,  3, 15,  4,  7, 12,  5,  6, 16,  8, 10, 14,  2},	// TITCHAINO		for Titchaino
	{ 17, 13,  9, 11,  3, 15,  4,  7, 12,  5,  6, 16,  8, 10, 14,  2},	// TITCHAINO_UNI	for Titchaino uni
	{ 17, 12, 13, 14,  5,  7,  3,  2,  9,  4,  8,  6, 16, 15, 11, 10},	// ONBOARD			for on board
	{  9,  4, 10,  6, 17, 11, 16, 13,  5, 15, 14,  8, 12,  7,  3,  2},	// DOTSDUINO		for Dotsduino
	{  2,  3,  4, 13, 14, 15, 16, 17, 12, 11, 10,  9,  5,  6,  7,  8},	// DOTSDUINO2D		for Dotsduino 2
	{ 15,  2, 13, 18,  5, 12,  6,  9, 19,  7,  8, 16, 11, 17,  3,  4},	// SQUARE8			for 8 Square
	{  3, 13,  9,  2, 11, 17,  5,  6, 21,  4,  7, 19,  8, 18, 14, 10},	// DOMINODUINO		for Dominoduino
};


/* -----------------------------------------------------------------  */
/** Table for the LED multiplexing cycles, containing 16 cycles made out of three bytes
 */
byte leds[2][16];

/// Determines whether the display is in single or double buffer mode
byte displayMode;

/// Flag indicating that the display page should be flipped as soon as the
/// current frame is displayed
boolean videoFlipPage;

/// Pointer to the buffer that is currently being displayed
byte* displayBuffer;

/// Pointer to the buffer that should currently be drawn to
byte* workBuffer;

static byte out_d = 0;
static byte out_c = 0;
static byte out_a = 0;
static byte out_f = 0;
static uint16_t sw1234;
static byte swdata;
static byte period;
static byte current_sw = 0;
static byte previous_sw = 0;
static byte scan;// 16 Cycles of Matrix
static byte msec;

#define MID_LED 6
#define MAX_LED 8


//### brightness equalization parameter
const int dd[MAX_LED + 1] PROGMEM= {1, 70, 120, 160, 190, 210, 220, 225 ,230};	//### duty ratio
//const int dd[MAX_LED] PROGMEM= {1, 180, 270, 370, 490, 650, 800, 950 };	//### duty ratio

// ADCœǂݎdlɂSW𔻕ʂ
const uint16_t th[4] = {117, 380, 650, 911}; // XbVhl
#define SWADC_PIN A16 // SWǂݎADCs
#define ACCX_PIN A4   // xZTX[q
#define ACCY_PIN A3   // xZTY[q
static Coordinate sensor; // xZTǂݎli[ϐ

/* -----------------------------------------------------------------  */
/** Table for LED Position in leds[] ram table 
 */

const byte *ledMap;

// all output Hi-Z
// LED쓮[qׂē͂ɂď
void alloff( void ){
#ifdef __AVR_ATmega328P__
	DDRD &= ~out_d; // line 76543210
	PORTD &= ~out_d;
	DDRB &= ~out_b; // line xxDCBA98
	PORTB &= ~out_b;
	DDRC &= ~out_c; // line xxJIHGFE
	PORTC &= ~out_c;
//	#ifndef titchaino_uni
//	DDRC &= ~0x30; // A4,A5
//	#endif
#else
	PORTA.DIRCLR = out_a; // PORTA 76543210
	PORTC.DIRCLR = out_c; // PORTC xxxxBA98
	PORTD.DIRCLR = out_d; // PORTD JIHGFEDC
	PORTF.DIRCLR = out_f; // PORTF xxxxxxLK
#endif
}

/* -----------------------------------------------------------------  */
/** The Interrupt code goes here ! 
    ^C}[荞݂LEDœ_i_Ci~bN_j
 */
#ifdef __AVR_ATmega328P__
// ATmega328P̏ꍇ
void scanLED(){
	alloff(); // all output high-Z (input)

	if ((scan==8)||(scan==9)){ // switch common lines
		PORTD |= in_d;	// pull up register of input port
		PORTB |= in_b;	// pull up register of input port
		PORTC |= in_c;	// pull up register of input port
	}

	byte n = pgm_read_byte(&ledMap[ scan ]); // scan column output LOW for cathode
	//byte n = ledMap[ scan ];
	if (n < 8) {
		DDRD  = _BV(n);
	} else if (n < 14 ){
		DDRB  = _BV(n-8);
	} else {
		DDRC  = _BV(n-14);
		//DDRC  = _BV(n-14) | 0x30; // A4,A5
	} 

	//### count ON LED numbers
	byte on_count = 0;	
	for (int j=0; j<8; j++){
		if (displayBuffer[scan*3] & 1<<j ){
		    on_count++;
		}
	}
	for (int j=0; j<6; j++){
		if (displayBuffer[scan*3+1] & 1<<j){
		    on_count++;
		}
		if (displayBuffer[scan*3+2] & 1<<j){
		    on_count++;
		}
	}

	if (scan==8){ // get SW1,SW2,SW3,SW4
		sw1234 = (digitalRead(sw1)*8 + digitalRead(sw2)*4 + digitalRead(sw3)*2 + digitalRead(sw4)) ^ 0x0F;
	} else if (scan==9){ // get SW5,SW6,SW7,SW8
		previous_sw = current_sw;
		if (assign == IKKEIPLEXING){
			current_sw = (sw1234 << 4) | ((digitalRead(sw5)*8 + digitalRead(sw6)*4 + digitalRead(sw7)*2 + digitalRead(sw8)) ^ 0x0F);
		} else {
			current_sw = ((digitalRead(sw9)<<1 | digitalRead(sw10)) ^ 0x03) << 6;
		}
		if (previous_sw == current_sw){
			swdata = current_sw;
		}
	} 
	
	// swdata: SW1,SW2,SW3,SW4,SW5,SW6,SW7,SW8
	if ((scan==8)||(scan==9)){ // switch common lines
		PORTD &= ~in_d;	// pull up register of input port
		PORTB &= ~in_b;	// pull up register of input port
		PORTC &= ~in_c;	// pull up register of input port
	}

	//### equalization pre-delay
	if ( (on_count >= MID_LED) && (on_count < MAX_LED) ){
		delayMicroseconds( 1000 - pgm_read_word_near(&dd[on_count]) );
	}

	// output row data
	DDRD  |= displayBuffer[scan*3];
	PORTD  = displayBuffer[scan*3];
	DDRB  |= displayBuffer[scan*3+1];
	PORTB  = displayBuffer[scan*3+1];
	DDRC  |= displayBuffer[scan*3+2];
	PORTC  = (PORTC & 0x30) | displayBuffer[scan*3+2];
#else
// AVR64DD28̏ꍇ
ISR(TCB0_INT_vect) { // ^C}[TCB0̊荞݃xNg
	TCB0.INTFLAGS = TCB_CAPT_bm; // ^C}[荞݃tOĂ

	alloff(); // all output high-Z (input)

	byte n = pgm_read_byte(&ledMap[ scan ]); // scan column output LOW for cathode
// _LED̃J\[hɂȂĂ[q̏o͒l炩߂ׂLɂĂ
	PORTA.OUTSET = out_a;
	PORTC.OUTSET = out_c;
	PORTD.OUTSET = out_d;
	PORTF.OUTSET = out_f;
// XL̓Am[hȂ̂ňHɂĂ
	if (n < 8) {
		PORTA.OUTCLR = _BV(n);
		PORTA.DIR = _BV(n);
	} else if (n < 12 ){
		PORTC.OUTCLR = _BV(n - 8);
		PORTC.DIR = _BV(n - 8);
	} else if (n < 20 ){
		PORTD.OUTCLR = _BV(n - 12);
		PORTD.DIR = _BV(n - 12);
	} else {
		PORTF.OUTCLR = _BV(n - 20);
		PORTF.DIR = _BV(n - 20);
	} 
// \obt@1񕪂̃f[^ǂݏo1̃rbgɑΉ[qo͂ɕςLED_
	byte data = displayBuffer[scan];
	for (int j=0; j<8; j++){
		if (data & _BV(j)){
			n = pgm_read_byte(&ledMap[ j + ((scan < 8)? 8:0)]);
			if (n < 8) {
				PORTA.DIRSET = _BV(n);
			} else if (n < 12 ){
				PORTC.DIRSET = _BV(n - 8);
			} else if (n < 20 ){
				PORTD.DIRSET = _BV(n - 12);
			} else {
				PORTF.DIRSET = _BV(n - 20);
			} 
		}
	}

// XL12̃^C~OŃXCb`̓dlADCœǂݎ
	if (scan==12){ // get SW1,SW2,SW3,SW4
		previous_sw = current_sw; // ÕXCb`ԍprevious_swɕۑ
		sw1234 = analogRead(SWADC_PIN); // ADCœǂݎXbVhlŃXCb`ԍ𔻕ʂ
		if (sw1234 > th[3]){
			current_sw = 0;
		} else if (sw1234 > th[2]){
			current_sw = 4;
		} else if (sw1234 > th[1]){
			current_sw = 3;
		} else if (sw1234 > th[0]){
			current_sw = 2;
		} else {
			current_sw = 1;
		}
		if (previous_sw == current_sw){ // OƓȂ炻XCb`ԍƂ
			swdata = current_sw;
		}
// XL4̃^C~OŉxZTǂݎ
	} else if (scan==4){
		sensor.x = analogRead(ACCX_PIN);
		sensor.y = analogRead(ACCY_PIN);
	}
	
//### count ON LED numbers
// 1̓_LED̐𐔂on_countɕۑ
	byte on_count = 0;
	for (int j=0; j<8; j++){
		if (data & 1<<j ) on_count++;
	}

//### equalization post-delay
// 1̓_LED̐ɉē_Ԃ𒲐
	delayMicroseconds( pgm_read_word_near(&dd[on_count]) );
	alloff();

#endif

// 10msTENMStO1ɂ
	period = 0;
	msec++;
	if (msec >= 10){
		msec = 0;
		period |= TENMS;
	}
// 1t[(16ms)FRAMEtO1ɂ
	scan++;
	if (scan >= 16){
		scan = 0;
		period |= FRAME;
		// If the page should be flipped, do it here.
		// _uobt@̏ꍇ1t[Ƀobt@ւ
		if (videoFlipPage && displayMode == DOUBLE_BUFFER){
		    // TODO: is this an atomic operation?
		    videoFlipPage = false;
		    byte* temp = displayBuffer;
		    displayBuffer = workBuffer;
		    workBuffer = temp;
		}
	}
}


/* -----------------------------------------------------------------  */
/** Constructor : Initialize the interrupt code. 
 * should be called in setup();
 */
void ikkeiplexing::Init(byte assign, byte mode){
// assign : IKKEIPLEXING, TITCHAINO, TITCHAINO_UNI, ONBOARD, DOTSDUINO, DOTSDUINO2D, SQUARE8, DOMINODUINO
	ledMap = maptable[assign];

#ifdef __AVR_ATmega328P__
	sw5 = sw1 = pgm_read_byte(&ledMap[12]);
	sw6 = sw2 = pgm_read_byte(&ledMap[13]);
	sw7 = sw3 = pgm_read_byte(&ledMap[14]);
	sw8 = sw4 = pgm_read_byte(&ledMap[15]);		

	
	if (assign == TITCHAINO_UNI){
		sw9  = 0;
		sw10 = 1;
	} else if (assign == SQUARE8){
		sw9  = 14; // A0
		sw10 = 14; // A0
	} else {
		sw9  = 18; // A4
		sw10 = 19; // A5
	}

	
	//### MsTimer2 setting
	MsTimer2::set(1,scanLED);
	MsTimer2::start();


	//### make using port bit pattern
	for (int j=0; j<16; j++){
		byte k = pgm_read_byte(&ledMap[ j ]);
		if ( k < 8 ){
			out_d |= _BV( k );      // line 76543210
		} else if ( k <14 ){
		    out_b |= _BV( k - 8 );  // line xxDCBA98
		} else {
		    out_c |= _BV( k - 14 ); // line xxJIHGFE
		}
	}
	
	//### make using input port bit pattern
	for (int j=12; j<16; j++){  // LA5,LA6,LA7,LA8
		byte k = pgm_read_byte(&ledMap[ j ]);
		if ( k < 8 ){
			in_d |= _BV( k );      // line 76543210
		} else if ( k <14 ){
		    in_b |= _BV( k - 8 );  // line xxDCBA98
		} else {
		    in_c |= _BV( k - 14 ); // line xxJIHGFE
		}
	}	

  	
	if (assign == TITCHAINO_UNI){
		in_d |= 0x03; 						// D0,D1 pullup
	} else if (assign != IKKEIPLEXING){
		in_c |= 0x30; 						// A4,A5 pullup
	} else if (assign == SQUARE8){
		in_c |= 0x01; 						// A0 pullup
	}
#else
// ^C}[TB0荞݂̐ݒ
// 1msɊ荞݂𔭐
	TCB0.CCMP = 24000 - 1; // 1/24MHz * 24000 = 1ms
	TCB0.CTRLB = 0;
	TCB0.INTCTRL = TCB_CAPT_bm;
	TCB0.CTRLA = TCB_CLKSEL_DIV1_gc | TCB_ENABLE_bm;

//### make using port bit pattern
// ledMap̃f[^rbgp^[
	for (int j=0; j<16; j++){
		byte k = pgm_read_byte(&ledMap[ j ]);
		if ( k < 8 ){
			out_a |= _BV( k );      // line 76543210
		} else if ( k <12 ){
			out_c |= _BV( k - 8 );  // line xxxxBA98
		} else if ( k <20 ){
		    out_d |= _BV( k - 12);  // line JIHGFEDC
		} else {
		    out_f |= _BV( k - 20 ); // line xxxxxxLK
		}
	}
#endif	

	// Record whether we are in single or double buffer mode
	displayMode = mode;

	// Point the display buffer to the first physical buffer
	displayBuffer = leds[0];

	// If we are in single buffered mode, point the work buffer
	// at the same physical buffer as the display buffer.  Otherwise,
	// point it at the second physical buffer.
	if( displayMode == SINGLE_BUFFER ) {
	    workBuffer = displayBuffer;
	} else {
	    workBuffer = leds[1];
	}

	scan = 0;
	msec = 0;

	// Clear the buffer and display it
	ikkeiplexing::Clear(0);
	ikkeiplexing::Flip(false);
}


/* -----------------------------------------------------------------  */
/* * Clear the screen completely
 * @param blocking if true : wait for flip before returning, if false :
 */
void ikkeiplexing::Flip(bool blocking){
    if (displayMode == DOUBLE_BUFFER){
        videoFlipPage = true;
        while (blocking && videoFlipPage) {
            delay(1);
        }
    }
}


/* -----------------------------------------------------------------  */
/** Clear the screen completely
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void ikkeiplexing::Clear(int set){
    for(int x=0;x<16;x++){
        for(int y=0;y<8;y++){
            Set(x,y,set);
        }
    }
}

/* -----------------------------------------------------------------  */
/** Shift left the screen completely
 */
void ikkeiplexing::ShiftLeft(){
    for(int x=1;x<16;x++){
        for(int y=0;y<8;y++){
            Set(x-1,y,Get(x,y));
        }
    }
}

/* -----------------------------------------------------------------  */
/** Shift right the screen completely
 */
void ikkeiplexing::ShiftRight(){
    for(int x=1;x<16;x++){
        for(int y=0;y<8;y++){
            Set(16-x,y,Get(15-x,y));
        }
    }
}

/* -----------------------------------------------------------------  */
/** Clear an horizontal line completely
 * @param y is the y coordinate of the line to clear/light [0-8]
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void ikkeiplexing::Horizontal(int y, int set){
    for(int x=0;x<16;x++){
        Set(x,y,set);
    }
}


/* -----------------------------------------------------------------  */
/** Clear a vertical line completely
 * @param x is the x coordinate of the line to clear/light [0-13]
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void ikkeiplexing::Vertical(int x, int set){
    for(int y=0;y<8;y++){
        Set(x,y,set);
    }
}

/* -----------------------------------------------------------------  */
void ikkeiplexing::Vwrite(int x, byte data){
    for(int y=0; y<8; y++){
    	if( data & (1<<y)){
    		Set(x, y, 1);
      	} else {
			Set(x, y, 0);
      	}
    }
}


/* -----------------------------------------------------------------  */
void ikkeiplexing::write(byte y, const uint16_t values[], byte size){
	for(byte i=0;i<size;i++){
		for(byte j=0;j<16;j++){
			if ((y+i) < 8){
				Set(j  , ((y+i) & 0x07), bitRead( values[i],15-j));
			} else {
				Set(j+8, ((y+i) & 0x07), bitRead( values[i],15-j));
			}
		}
	}
}

/* -----------------------------------------------------------------  */
void ikkeiplexing::write90(byte y, const uint16_t values[], byte size){
	for(byte i=0;i<size;i++){
		for(byte j=0;j<16;j++){
			Set(((y+i) & 0x0F), j, bitRead( values[i],15-j));
		}
	}
}

/* -----------------------------------------------------------------  */
void ikkeiplexing::Set(byte x, byte y, byte c){
	if (c == 1) {
		workBuffer[x] |=  _BV(y);
	} else {
		workBuffer[x] &= ~_BV(y);
	}
	
}

/* -----------------------------------------------------------------  */
byte ikkeiplexing::Get(byte x, byte y){
	return ((workBuffer[x] & _BV(y)) != 0);
}


/* -----------------------------------------------------------------  */
/** GetSW : get switch status
    0:SW_LEVEL (level)
    1:SW_PEDGE (positive edge)
    2:SW_NEDGE (negative edge)
 */
byte ikkeiplexing::GetSW(byte edge){
	static byte previous_swdata = 0;
	byte flag = 0;
#ifdef __AVR_ATmega328P__
	if (edge == 0){
		flag = swdata;
	} else if (edge == 1){
		flag = swdata & ~previous_swdata;
	} else if (edge == 2){
		flag = ~swdata & previous_swdata;
	}
#else		
	if (edge == 0){
		flag = swdata;
	} else if (edge == 1){
		if (swdata != previous_swdata){
			flag = swdata;
		}
	} else if (edge == 2){
		if (swdata != previous_swdata){
			flag = previous_swdata;
		}
	}
#endif
	previous_swdata = swdata;
 	return (flag);
} 


/* -----------------------------------------------------------------  */
/** GetPeriod : get period flag
 */
byte ikkeiplexing::GetPeriod(byte cbit){
	digitalWrite(PIN_PA0, 0); // dummy omajinai
	
 	byte flag = period & cbit;
	period &= ~cbit;
 	return (flag);
} 


/* -----------------------------------------------------------------  */
/** GetAcceleration : get acceleration
 */
Coordinate ikkeiplexing::GetAcceleration(){
	return (sensor);
}

